//===----- llvm/IR/DFGraph.h - Classes to represent a Dataflow Graph ------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file contains the definition of the following classes: 
// 1. DFNode
// 2. DFGraph
// 3. DFInternalNode
// 4. DFLeafNode
// 5. DFEdge.
//
// FIXME : We still need to figure out whether these functions are independent
//===----------------------------------------------------------------------===//

#ifndef LLVM_IR_DFGRAPH_H
#define LLVM_IR_DFGRAPH_H

#include "llvm/IR/Function.h"
#include "llvm/IR/IntrinsicInst.h"
#include "llvm/IR/Value.h"
#include "llvm/Support/Compiler.h"
#include "llvm/ADT/GraphTraits.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/GraphWriter.h"


namespace llvm {

class DFNode;
class DFInternalNode;
class DFLeafNode;
class DFEdge;
class DFNodeVisitor;
class DFTreeTraversal;
class DFEdgeVisitor;
class DFGraph;

//template<> struct GraphTraits

typedef std::vector<DFNode*> DFNodeListType;

class DFGraph {

private:
  typedef std::vector<DFEdge*> DFEdgeListType;

  // Important things that make up a Dataflow graph
  //  DFLeafNode* Entry;
  DFInternalNode* Parent;
  DFNodeListType ChildrenList;    ///< List of children Dataflow Nodes
  DFEdgeListType DFEdgeList;      ///< List of Dataflow edges among children

public:
  DFGraph(DFInternalNode* P) {
    //ChildrenList.push_back(llvm::DFLeafNode::Create(NULL, NULL, NULL));
    Parent = P;
  }

  void addChildDFNode(DFNode* child) {
    ChildrenList.push_back(child);
  }

  void addDFEdge(DFEdge* E) {
    DFEdgeList.push_back(E);
  }

  // Iterators
  typedef DFNodeListType::iterator children_iterator;
  typedef DFNodeListType::const_iterator const_children_iterator;

  typedef DFEdgeListType::iterator dfedge_iterator;
  typedef DFEdgeListType::const_iterator const_dfedge_iterator;

  //===--------------------------------------------------------------------===//
  // DFNodeList iterator forwarding functions
  //
  children_iterator       begin()       { return ChildrenList.begin(); }
  const_children_iterator begin() const { return ChildrenList.begin(); }
  children_iterator       end  ()       { return ChildrenList.end();   }
  const_children_iterator end  () const { return ChildrenList.end();   }

  size_t                   size() const { return ChildrenList.size();  }
  bool                    empty() const { return ChildrenList.empty(); }
  const DFNode           *front() const { return ChildrenList.front(); }
        DFNode           *front()       { return ChildrenList.front(); }
  const DFNode            *back() const { return ChildrenList.back();  }
        DFNode            *back()       { return ChildrenList.back();  }

  //===--------------------------------------------------------------------===//

  //===--------------------------------------------------------------------===//
  // DFEdgeList iterator forwarding functions
  //
  dfedge_iterator       dfedge_begin()       { return DFEdgeList.begin(); }
  const_dfedge_iterator dfedge_begin() const { return DFEdgeList.begin(); }
  dfedge_iterator       dfedge_end  ()       { return DFEdgeList.end();   }
  const_dfedge_iterator dfedge_end  () const { return DFEdgeList.end();   }

  size_t                 dfedge_size() const { return DFEdgeList.size();  }
  bool                  dfedge_empty() const { return DFEdgeList.empty(); }
  const DFEdge         *dfedge_front() const { return DFEdgeList.front(); }
        DFEdge         *dfedge_front()       { return DFEdgeList.front(); }
  const DFEdge          *dfedge_back() const { return DFEdgeList.back();  }
        DFEdge          *dfedge_back()       { return DFEdgeList.back();  }

  //===--------------------------------------------------------------------===//

  DFInternalNode* getParent() {
    return Parent;
  }

};

// DFNode represents a single VISC Dataflow Node in LLVM.
// This is an abstract class.
//
// A Dataflow Node basically consists of
// 1. Pointer to a function describing this dataflow node
// 2. Number of dimensions in which the node is replicated
// 3. Number of instances in each dimension
// 4. Pointer to parent Dataflow Node
// 5. List of children Dataflow Nodes (empty if it is a leaf node)
// 6. List of Dataflow Edges among children

class DFNode {

  public:
  enum DFNodeKind {
    Internal,
    Leaf
  };

  private:

  const DFNodeKind Kind;

  // Important things that make up a Dataflow Node
  IntrinsicInst* II;              ///< Associated IntrinsicInst/Value
  Function* FuncPointer;          ///< Associated Function
  DFNode* Parent;                 ///< Pointer to parent dataflow Node
  int NumOfDim;                   ///< Number of dimensions
  std::vector<Value*> DimLimits;  ///< Number of instances in each dimension
  DFNodeListType Successors;      ///< List of successors i.e.,
                                  ///< destination DFNodes to DFEdges
                                  ///< originating from this DFNode

  public:
  DFNodeKind getKind() const {return Kind;}

  // Iterators
  typedef DFNodeListType::iterator successor_iterator;
  typedef DFNodeListType::const_iterator const_successor_iterator;

  //===--------------------------------------------------------------------===//
  // DFNodeList iterator forwarding functions
  //
  successor_iterator       successors_begin()        { return Successors.begin(); }
  const_successor_iterator successors_begin()  const { return Successors.begin(); }
  successor_iterator       successors_end  ()        { return Successors.end();   }
  const_successor_iterator successors_end  ()  const { return Successors.end();   }

  size_t                   successors_size()   const { return Successors.size();  }
  bool                     successors_empty()  const { return Successors.empty(); }
  const DFNode*            successors_front()  const { return Successors.front(); }
        DFNode*            successors_front()        { return Successors.front(); }
  const DFNode*            successors_back()   const { return Successors.back();  }
        DFNode*            successors_back()         { return Successors.back();  }

  //===--------------------------------------------------------------------===//

  // Functions
  DFNode(DFNodeKind _Kind, IntrinsicInst* _II, Function* _FuncPointer, DFNode* _Parent,
         int _NumOfDim, std::vector<Value*> _DimLimits) : Kind(_Kind), II(_II),
         FuncPointer(_FuncPointer), Parent(_Parent), NumOfDim(_NumOfDim),
         DimLimits(_DimLimits) {}

  void addSuccessor(DFNode* N) {
    Successors.push_back(N);
  }

  Function* getFuncPointer() {
    return FuncPointer;
  }



  virtual void applyDFNodeVisitor(DFNodeVisitor &V, DFNodeListType *L = NULL) = 0;
//  virtual void applyDFEdgeVisitor(DFEdgeVisitor &V) = 0;

};


class DFInternalNode : public DFNode {

  private:
  DFGraph* childGraph;

  // Constructor
  DFInternalNode(IntrinsicInst* II, Function* FuncPointer, DFNode* Parent,
                 int NumOfDim, std::vector<Value*> DimLimits) :
                 DFNode(Internal, II, FuncPointer, Parent, NumOfDim, DimLimits) {
    childGraph = new DFGraph(this);
    //childGraph->addChildDFNode(DFLeafNode::Create(NULL, NULL, this));
  }

  public:
  static DFInternalNode *Create(IntrinsicInst* II, Function* FuncPointer,
                                DFNode* Parent = NULL, int NumOfDim = 0,
                                std::vector<Value*> DimLimits = std::vector<Value*>()) {
    return new DFInternalNode(II, FuncPointer, Parent, NumOfDim, DimLimits);
  }

  static bool classof(const DFNode *N) {
    return N->getKind() == Internal;
  }
 

  void addChildToDFGraph(DFNode* N) {
    childGraph->addChildDFNode(N);
  }

  void addEdgeToDFGraph(DFEdge* E) {
    childGraph->addDFEdge(E);
  }

  DFGraph* getChildGraph() {
    return childGraph;
  }

  void applyDFNodeVisitor(DFNodeVisitor &V, DFNodeListType *L = NULL); /*virtual*/
//  void applyDFEdgeVisitor(DFEdgeVisitor &V); /*virtual*/
};

class DFLeafNode : public DFNode {
  private:
  // Constructor
  DFLeafNode(IntrinsicInst* II, Function* FuncPointer, DFNode* Parent,
             int NumOfDim = 0, std::vector<Value*> DimLimits = std::vector<Value*>())
            : DFNode(Leaf, II, FuncPointer, Parent, NumOfDim, DimLimits) {}

  public:

  static DFLeafNode *Create(IntrinsicInst* II, Function* FuncPointer,
                            DFNode* Parent, int NumOfDim = 0,
                            std::vector<Value*> DimLimits = std::vector<Value*>()) {
    return new DFLeafNode(II, FuncPointer, Parent, NumOfDim, DimLimits);
  }

  static bool classof(const DFNode *N) {
    return N->getKind() == Leaf;
  }


  void applyDFNodeVisitor(DFNodeVisitor &V, DFNodeListType *L = NULL); /*virtual*/
//  void applyDFEdgeVisitor(DFEdgeVisitor &V); /*virtual*/

};

// DFEdge represents a single VISC Dataflow Edge in LLVM.
//
// A Dataflow Edge basically consists of
// 1. Pointer to the dataflow node that is the source of this edge
// 2. Pointer to the dataflow node that is the destination of this edge
// 3. Pointer to a function that describes which instances of the source
//    dataflow node are connected to which instances of the destination
//    dataflow node via this edge
// 4. Pointer to a function that describes which input arguments of the
//    destination dataflow node are connected to which outputs of the source
//    dataflow node via this edge

class DFEdge {
  private:
  // Important things that make up a Dataflow Edge
  DFNode* SrcDF;                ///< Pointer to source dataflow Node
  DFNode* DestDF;               ///< Pointer to destination dataflow Node
  Function* DFMapFuncPointer;   ///< Function that associates the appropriate
                                ///< instances of source and destination
                                ///< dataflow nodes
  Function* ArgMapFuncPointer;  ///< Function that associates the input
                                ///< arguments of destination with the outputs
                                ///< of source dataflow node
  // Functions
  DFEdge(DFNode* _SrcDF, DFNode* _DestDF, Function* _DFMapFuncPointer,
         Function* _ArgMapFuncPointer) : SrcDF(_SrcDF), DestDF(_DestDF),
         DFMapFuncPointer(_DFMapFuncPointer),
         ArgMapFuncPointer(_ArgMapFuncPointer) {}

  public:

  static DFEdge *Create(DFNode* SrcDF, DFNode* DestDF, Function* DFMapFuncPtr,
                        Function* ArgMapFuncPtr) {
    return new DFEdge(SrcDF, DestDF, DFMapFuncPtr, ArgMapFuncPtr);

  }
};


//===-------------------------- Visitor Classes ---------------------------===//
// Visitor for DFNode objects
class DFNodeVisitor {
  public:
  virtual void visit(DFInternalNode* N, DFNodeListType* L = NULL) = 0;
  virtual void visit(DFLeafNode* N, DFNodeListType* L = NULL) = 0;
};

class DFTreeTraversal : public DFNodeVisitor {

  public:
  virtual void visit(DFInternalNode* N, DFNodeListType *L = NULL){
    errs() << "Visted Node (I) - " << N->getFuncPointer()->getName() << "\n";
    if (L != NULL)
      L->push_back(N);
    for(DFGraph::children_iterator i = N->getChildGraph()->begin(),
        e = N->getChildGraph()->end(); i != e; ++i) {
      DFNode* child = *i;
      child->applyDFNodeVisitor(*this, L);
    }
  }

  virtual void visit(DFLeafNode* N, DFNodeListType *L = NULL) {
    errs() << "Visted Node (L) - " << N->getFuncPointer()->getName() << "\n";
    if (L != NULL)
      L->push_back(N);
  }

};

class FollowSuccessors : public DFNodeVisitor {

  public:
  virtual void visit(DFInternalNode* N, DFNodeListType *L = NULL) {
    errs() << "Visted Node (I) - " << N->getFuncPointer()->getName() << "\n";
    for(DFInternalNode::successor_iterator i = N->successors_begin(),
        e = N->successors_end(); i != e; ++i) {
      /* Traverse the graph.
       * Choose the kind of traversal we want
       * Do we do a DAG kind of traversal?
       */
    }
  }

  virtual void visit(DFLeafNode* N, DFNodeListType* L = NULL) {
    errs() << "Visted Node (L) - " << N->getFuncPointer()->getName() << "\n";
  }
};

// Print functions
inline raw_ostream& operator<<(raw_ostream &O, DFInternalNode &N) {
    O << N.getFuncPointer()->getName();
    return O;
}

inline raw_ostream& operator<<(raw_ostream &O, DFLeafNode &N) {
    O << N.getFuncPointer()->getName();
    return O;
}

/*
// Visitor for DFEdge objects
class DFEdgeVisitor {
public:
  virtual void visit(DFEdge* E) = 0;
};


//===--------------------------------------------------------------------===//
// GraphTraits specializations for DFNode graph (DFG)
//===--------------------------------------------------------------------===//

// Provide specializations of GraphTraits to be able to treat a DFNode as a
// graph of DFNodes...struct GraphTraits {
  // Elements to provide:

  // typedef NodeType          - Type of Node in the graph
  // typedef ChildIteratorType - Type used to iterate over children in graph

  // static NodeType *getEntryNode(const GraphType &)
  //    Return the entry node of the graph

  // static ChildIteratorType child_begin(NodeType *)
  // static ChildIteratorType child_end  (NodeType *)
  //    Return iterators that point to the beginning and ending of the child
  //    node list for the specified node.
  //


  // typedef  ...iterator nodes_iterator;
  // static nodes_iterator nodes_begin(GraphType *G)
  // static nodes_iterator nodes_end  (GraphType *G)
  //    nodes_iterator/begin/end - Allow iteration over all nodes in the graph

  // static unsigned       size       (GraphType *G)
  //    Return total number of nodes in the graph
  //


  // If anyone tries to use this class without having an appropriate
  // specialization, make an error.  If you get this error, it's because you
  // need to include the appropriate specialization of GraphTraits<> for your
  // graph, or you need to define it for a new graph type. Either that or
  // your argument to XXX_begin(...) is unknown or needs to have the proper .h
  // file #include'd.
  //
  typedef typename GraphType::UnknownGraphTypeError NodeType;
//};
*/

} // End llvm namespace

#endif
