//
//===---- DFG2LLVM.h - Header file for "VISC Dataflow Graph to Target" ----===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

#ifndef VISC_UTILS_HEADER
#define VISC_UTILS_HEADER

#include <assert.h>
 
#include "llvm/IR/Module.h"
#include "llvm/IR/Function.h"
#include "llvm/IR/InstIterator.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/IntrinsicInst.h"
#include "llvm/IR/Value.h"
#include "llvm/Pass.h"
#include "llvm/IR/Metadata.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/SupportVISC/VISCHint.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Transforms/Utils/ValueMapper.h"

using namespace llvm;

namespace viscUtils {
// Helper Functions

static bool isViscCreateNodeIntrinsic(Instruction* I) {
  if(!isa<IntrinsicInst>(I))
    return false;
  IntrinsicInst* II = cast<IntrinsicInst>(I);
  return (II->getCalledFunction()->getName()).startswith("llvm.visc.createNode");
}

static bool isViscCreateNodeCall(Instruction* I) {
  if(!isa<CallInst>(I))
    return false;
  CallInst* CI = cast<CallInst>(I);
  return (CI->getCalledValue()->stripPointerCasts()->getName()).startswith("__visc__createNode");
}

static bool isViscLaunchCall(Instruction* I) {
  if(!isa<CallInst>(I))
    return false;
  CallInst* CI = cast<CallInst>(I);
  return (CI->getCalledValue()->stripPointerCasts()->getName()).startswith("__visc__launch");
}
// Creates a new createNode intrinsic, similar to II but with different
// associated function F instead
IntrinsicInst* createIdenticalCreateNodeIntrinsicWithDifferentFunction(Function* F,
                                                                       IntrinsicInst* II) {
  Module* M = F->getParent();

  // Find which createNode intrinsic we need to create
  Function* CreateNodeF = Intrinsic::getDeclaration(M, II->getIntrinsicID());
  Constant* Fp = ConstantExpr::getPointerCast(F,
                                          Type::getInt8PtrTy(II->getContext()));

  ArrayRef<Value*> CreateNodeArgs;
  switch (II->getIntrinsicID()) {
    case Intrinsic::visc_createNode:
    {
      CreateNodeArgs = ArrayRef<Value*>(Fp);
      break;
    }
    case Intrinsic::visc_createNode1D:
    {
      Value* CreateNode1DArgs[] = {Fp, II->getArgOperand(1)};
      CreateNodeArgs = ArrayRef<Value*>(CreateNode1DArgs, 2);
      break;
    }
    case Intrinsic::visc_createNode2D:
    {
      Value* CreateNode2DArgs[] = {Fp, II->getArgOperand(1),
                                       II->getArgOperand(2)};
      CreateNodeArgs = ArrayRef<Value*>(CreateNode2DArgs, 3);
      break;
    }
    case Intrinsic::visc_createNode3D:
    {
      Value* CreateNode3DArgs[] = {Fp, II->getArgOperand(1),
                                       II->getArgOperand(2),
                                       II->getArgOperand(3)};
      CreateNodeArgs = ArrayRef<Value*>(CreateNode3DArgs, 4);
      break;
    }
    default :
      assert(false && "Unknown createNode intrinsic");
      break;
  }

  CallInst* CI = CallInst::Create(CreateNodeF,
                                  CreateNodeArgs,
                                  F->getName()+".node");
  IntrinsicInst* CreateNodeII = cast<IntrinsicInst>(CI);
  return CreateNodeII;
}
/*
CallInst* createIdenticalCreateNodeCallWithDifferentFunction(Function* F,
                                                             CallInst* CI) {

  // Find which createNode function call we need to create
  Function* CreateNodeF = CI->getCalledValue();

  ArrayRef<Value*> CreateNodeArgs;
  if ((CreateNodeF->stripPointerCasts()->getName()).equals("__visc__createNode")) {
    // This is a createNode call
    CreateNodeArgs = ArrayRef<Value*>(CreateNodeF);
  } else if ((CreateNodeF->stripPointerCasts()->getName()).equals("__visc__createNode1D")) {
    // This is a createNode1D call
    Value* CreateNode1DArgs[] = {CreateNodeF, CI->getArgOperand(1)};
    CreateNodeArgs = ArrayRef<Value*>(CreateNode1DArgs, 2);
  } else if ((CreateNodeF->stripPointerCasts()->getName()).equals("__visc__createNode2D")) {
    // This is a createNode2D call
    Value* CreateNode2DArgs[] = {CreateNodeF,
                                 CI->getArgOperand(1),
                                 CI->getArgOperand(2)};
    CreateNodeArgs = ArrayRef<Value*>(CreateNode2DArgs, 3);
  } else if ((CreateNodeF->stripPointerCasts()->getName()).equals("__visc__createNode3D")) {
    // This is a createNode3D call
    Value* CreateNode3DArgs[] = {CreateNodeF,
                                 CI->getArgOperand(1),
                                 CI->getArgOperand(2),
                                 CI->getArgOperand(3)};
    CreateNodeArgs = ArrayRef<Value*>(CreateNode3DArgs, 4);
  } else {
    assert(false && "Unknown createNode call");
  }

  CallInst* newCI = CallInst::Create(CreateNodeF,
                                     CreateNodeArgs,
                                     F->getName()+".cncall");
  return newCI;
}
*/

// Fix VISC hints for this function
void fixHintMetadata(Module &M, Function* F, Function* G) {
    Metadata* MD_F = ValueAsMetadata::getIfExists(F);
    MDTuple* MDT_F = MDTuple::getIfExists(F->getContext(), ArrayRef<Metadata*>(MD_F));
    DEBUG(errs() << "Associated Metadata: " << *MDT_F << "\n");
    MDTuple* MDT_G = MDNode::get(F->getContext(), ArrayRef<Metadata*>(ValueAsMetadata::get(G)));
    DEBUG(errs() << "New Metadata: " << *MDT_G << "\n");

    NamedMDNode* HintNode = M.getOrInsertNamedMetadata("visc_hint_gpu");
    for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
        if(HintNode->getOperand(i) == MDT_F)
           HintNode->setOperand(i, MDT_G);
    }
    HintNode = M.getOrInsertNamedMetadata("visc_hint_spir");
    for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
        if(HintNode->getOperand(i) == MDT_F)
           HintNode->setOperand(i, MDT_G);
    }
    HintNode = M.getOrInsertNamedMetadata("visc_hint_cudnn");
    for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
        if(HintNode->getOperand(i) == MDT_F)
           HintNode->setOperand(i, MDT_G);
    }
    HintNode = M.getOrInsertNamedMetadata("visc_hint_promise");
    for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
        if(HintNode->getOperand(i) == MDT_F)
           HintNode->setOperand(i, MDT_G);
    }
    HintNode = M.getOrInsertNamedMetadata("visc_hint_cpu");
    for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
        if(HintNode->getOperand(i) == MDT_F)
           HintNode->setOperand(i, MDT_G);
    }
    HintNode = M.getOrInsertNamedMetadata("visc_hint_cpu_gpu");
    for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
        if(HintNode->getOperand(i) == MDT_F)
           HintNode->setOperand(i, MDT_G);
    }
    HintNode = M.getOrInsertNamedMetadata("visc_hint_cpu_spir");
    for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
        if(HintNode->getOperand(i) == MDT_F)
           HintNode->setOperand(i, MDT_G);
    }
}

// Assuming that the changed function is a node function, it is only used as a
// first operand of createNode*. It is enough to iterate through all createNode*
// calls in the program.
void replaceNodeFunctionInIR(Module &M, Function* F, Function* G) {

  for (Module::iterator mi = M.begin(), me = M.end(); mi != me; ++mi) {
    Function* f = &*mi;
    DEBUG(errs() << "Function: " << f->getName() << "\n");

    std::vector<Instruction*> toBeErased;

    for (inst_iterator i = inst_begin(f), e = inst_end(f); i != e ; ++i) {
      Instruction* I = &*i; // Grab pointer to Instruction

      if (isViscCreateNodeIntrinsic(I)) {
        IntrinsicInst* II = cast<IntrinsicInst>(I);
        // The found createNode is not associated with the changed function
        if (II->getArgOperand(0) != F)
          continue; // skip it

        // Otherwise, create a new createNode similar to the other one,
        // but with the changed function as first operand.
        IntrinsicInst* CreateNodeII =
          createIdenticalCreateNodeIntrinsicWithDifferentFunction(G, II);
        II->replaceAllUsesWith(CreateNodeII);
        toBeErased.push_back(II);
      } else if (isViscCreateNodeCall(I)) {
        CallInst* CI = cast<CallInst>(I);
        // The found createNode is not associated with the changed function
        if (CI->getArgOperand(1) != F)
          continue; // skip it

        DEBUG(errs() << "Fixing use: " << *CI << "\n");
        DEBUG(errs() << "in function: " << f->getName() << "\n");
        // Replace use of F with use of G
        CI->setArgOperand(1, G);
        DEBUG(errs() << "Fixed use: " << *CI << "\n");
      } else if(isViscLaunchCall(I)) {
        CallInst* CI = cast<CallInst>(I);
        // The found launch call is not associated with the changed function
        if (CI->getArgOperand(1)->stripPointerCasts() != F)
          continue;

        // Otherwise, replace F with G
        errs() << *G->getType() << "\n";
        errs() << *CI->getArgOperand(1)->getType() << "\n";
        CI->setArgOperand(1, G);
      }

    }

    for(auto I: toBeErased) {
      DEBUG(errs() << "\tErasing " << *I << "\n");
      I->eraseFromParent();
    }
  }

  // Check if the function is used by a metadata node
  if(F->isUsedByMetadata()) {
    fixHintMetadata(M, F, G);
  }
  DEBUG(errs() << "DONE: Replacing function " << F->getName() << " with " << G->getName() << "\n");

  // Remove replaced function from the module
  //assert(F->user_empty() && "Still some uses of older function left\n");
  F->replaceAllUsesWith(UndefValue::get(F->getType()));
  F->eraseFromParent();

}


// Create new function F' as a copy of old function F with a new signature.
// The following two most used cases are handled by this function.
// 1. When some extra arguments need to be added to this function
//    - Here we can map the old function arguments to
//      new ones
// 2. When each pointer argument needs an additional size argument
//    - Here, in the absence of VMap, we map the arguments in order, skipping
//      over extra pointer arguments.
// The function returns the list of return instructions to the caller to fix in
// case the return type is also changed.
Function* cloneFunction(Function* F, FunctionType* newFT, bool
    isAddingPtrSizeArg, SmallVectorImpl<ReturnInst*>* Returns = NULL) {

  DEBUG(errs() << "Cloning Function: " << F->getName() << "\n");
  DEBUG(errs() << "Old Function Type: " << *F->getFunctionType() << "\n");
  DEBUG(errs() << "New Function Type: " << *newFT << "\n");

  assert(F->getFunctionType()->getNumParams() <= newFT->getNumParams()
      && "This function assumes that the new function has more arguments than the old function!");

  // Create Function of specified type
  Function* newF = Function::Create(newFT, F->getLinkage(), F->getName()+"_cloned", F->getParent());
  DEBUG(errs() << "Old Function name: " << F->getName() << "\n");
  DEBUG(errs() << "New Function name: " << newF->getName() << "\n");
  ValueToValueMapTy VMap;
  DEBUG(errs() << "No value map provided. Creating default value map\n");
  if(isAddingPtrSizeArg) {
    DEBUG(errs() << "Case 1: Pointer arg followed by a i64 size argument in new function\n");
    Function::arg_iterator new_ai = newF->arg_begin();
    for(Function::arg_iterator ai = F->arg_begin(), ae = F->arg_end();
        ai != ae; ++ai) {
      DEBUG(errs() << ai->getArgNo() << ". " << *ai << " : " << *new_ai << "\n");
      assert(ai->getType() == new_ai->getType() && "Arguments type do not match!");
      VMap[&*ai] = &*new_ai;
      new_ai->takeName(&*ai);
      if(ai->getType()->isPointerTy()) {
        std::string oldName = new_ai->getName();
        // If the current argument is pointer type, the next argument in new
        // function would be an i64 type containing the data size of this
        // argument. Hence, skip the next arguement in new function.
        ++new_ai;
        new_ai->setName("bytes_"+oldName);
      }
      ++new_ai;
    }
  }
  else {
    DEBUG(errs() << "Case 2: Extra arguments are added at the end of old function\n");
    Function::arg_iterator new_ai = newF->arg_begin();
    for(Function::arg_iterator ai = F->arg_begin(), ae = F->arg_end();
        ai != ae; ++ai, ++new_ai) {
      DEBUG(errs() << ai->getArgNo() << ". " << *ai << " : " << *new_ai << "\n");
      assert(ai->getType() == new_ai->getType() && "Arguments type do not match!");
      VMap[&*ai] = &*new_ai;
      new_ai->takeName(&*ai);
    }
  }

  // Clone function
  if (Returns == NULL)
    Returns = new SmallVector<ReturnInst*, 8>();
  CloneFunctionInto(newF, F, VMap, false, *Returns);

  return newF;
}

 //------------------- Helper Functions For Handling Hints -------------------//
  
// Return true if 1st arg (tag) contains 2nd (target)
bool tagIncludesTarget(visc::Target Tag, visc::Target T) {
  switch (Tag) {
    case visc::None:
      return false;
    case visc::CPU_TARGET:
      if (T == visc::CPU_TARGET)
        return true;
      else
        return false;
    case visc::GPU_TARGET:
      if (T == visc::GPU_TARGET)
        return true;
      else
        return false;
    case visc::SPIR_TARGET:
      if (T == visc::SPIR_TARGET)
        return true;
      else
        return false;
    case visc::CUDNN_TARGET:
      if (T == visc::CUDNN_TARGET)
        return true;
      else
        return false;
    case visc::PROMISE_TARGET:
      if (T == visc::PROMISE_TARGET)
        return true;
      else
        return false;
    case visc::CPU_OR_GPU_TARGET:
      if ((T == visc::CPU_TARGET) ||
          (T == visc::GPU_TARGET) ||
          (T == visc::CPU_OR_GPU_TARGET))
        return true;
      else
        return false;
    case visc::CPU_OR_SPIR_TARGET:
      if ((T == visc::CPU_TARGET) ||
          (T == visc::SPIR_TARGET) ||
          (T == visc::CPU_OR_SPIR_TARGET))
        return true;
      else
        return false;
    default:
      assert(false && "Unknown Target\n");
  }
}

bool isSingleTargetTag(visc::Target T) {
  return ((T == visc::CPU_TARGET)    ||
          (T == visc::GPU_TARGET)    ||
          (T == visc::SPIR_TARGET)   ||
          (T == visc::CUDNN_TARGET)  ||
          (T == visc::PROMISE_TARGET));
}

// Add the specified target to the given tag
visc::Target getUpdatedTag(visc::Target Tag, visc::Target T) {
  assert(((T == visc::CPU_TARGET)    ||
          (T == visc::GPU_TARGET)    ||
          (T == visc::SPIR_TARGET)   ||
          (T == visc::CUDNN_TARGET)  ||
          (T == visc::PROMISE_TARGET)) &&
         "The target is only allowed to be a single target: CPU, GPU, SPIR, CUDNN, PROMISE\n");

  switch (Tag) {
    case visc::None:
      return T;
    case visc::CPU_TARGET:
      assert((T != visc::CUDNN_TARGET) && (T != visc::PROMISE_TARGET) &&
             "Unsupported target combination\n");
      break;
      if (T == visc::CPU_TARGET)
        return visc::CPU_TARGET;
      if (T == visc::GPU_TARGET)
        return visc::CPU_OR_GPU_TARGET;
      if (T == visc::SPIR_TARGET)
        return visc::CPU_OR_SPIR_TARGET;
    case visc::GPU_TARGET:
      assert((T != visc::SPIR_TARGET) && "Unsupported target combination\n");
      assert((T != visc::CUDNN_TARGET) && (T != visc::PROMISE_TARGET) &&
             "Unsupported target combination\n");
      break;
      if (T == visc::CPU_TARGET)
        return visc::CPU_OR_GPU_TARGET;
      if (T == visc::GPU_TARGET)
        return visc::GPU_TARGET;
    case visc::SPIR_TARGET:
      assert((T != visc::GPU_TARGET) && "Unsupported target combination\n");
      assert((T != visc::CUDNN_TARGET) && (T != visc::PROMISE_TARGET) &&
             "Unsupported target combination\n");
      break;
      if (T == visc::CPU_TARGET)
        return visc::CPU_OR_SPIR_TARGET;
      if (T == visc::SPIR_TARGET)
        return visc::SPIR_TARGET;
    case visc::CPU_OR_GPU_TARGET:
      assert((T != visc::CUDNN_TARGET) && (T != visc::PROMISE_TARGET) &&
             "Unsupported target combination\n");
      break;
      assert((T != visc::SPIR_TARGET) && "Unsupported target combination\n");
      return visc::CPU_OR_GPU_TARGET;
    case visc::CPU_OR_SPIR_TARGET:
      assert((T != visc::CUDNN_TARGET) && (T != visc::PROMISE_TARGET) &&
             "Unsupported target combination\n");
      break;
      assert((T != visc::GPU_TARGET) && "Unsupported target combination\n");
      return visc::CPU_OR_SPIR_TARGET;
    default:
      assert(false && "Unknown Target\n");
  }
}

// This functions add the hint as metadata in visc code
void addHint(Function* F, visc::Target T) {
   // Get Module
  Module* M = F->getParent();
  DEBUG(errs() << "Set preferred target for " << F->getName() << ": ");
 
  // Based on the hint, get the hint metadata
  NamedMDNode* HintNode;
  switch (T) {
    case visc::GPU_TARGET:
      DEBUG(errs() << "GPU Target\n");
      HintNode = M->getOrInsertNamedMetadata("visc_hint_gpu");
      break;
    case visc::SPIR_TARGET:
      DEBUG(errs() << "SPIR Target\n");
      HintNode = M->getOrInsertNamedMetadata("visc_hint_spir");
      break;
    case visc::CUDNN_TARGET:
      DEBUG(errs() << "CUDNN Target\n");
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cudnn");
      break;
    case visc::PROMISE_TARGET:
      DEBUG(errs() << "PROMISE Target\n");
      HintNode = M->getOrInsertNamedMetadata("visc_hint_promise");
      break;
    case visc::CPU_TARGET:
      DEBUG(errs() << "CPU Target\n");
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu");
      break;
    case visc::CPU_OR_GPU_TARGET:
      DEBUG(errs() << "CPU or GPU Target\n");
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu_gpu");
      break;
    case visc::CPU_OR_SPIR_TARGET:
      DEBUG(errs() << "CPU or SPIR Target\n");
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu_spir");
      break;
    default:
      llvm_unreachable("Unsupported Target Hint!");
      break;
  }

  // Create a node for the function and add it to the hint node
  MDTuple* N = MDNode::get(M->getContext(), ArrayRef<Metadata*>(ValueAsMetadata::get(F)));
  HintNode->addOperand(N);
}

// This function removes the hint as metadata in visc code
void removeHint(Function* F, visc::Target T) {
  // Get Module
  Module* M = F->getParent();
  DEBUG(errs() << "Remove preferred target for " << F->getName() << ": " << T << "\n");

  // Based on the hint, get the hint metadata
  NamedMDNode* HintNode;
  switch (T) {
    case visc::GPU_TARGET:
      HintNode = M->getOrInsertNamedMetadata("visc_hint_gpu");
      break;
    case visc::SPIR_TARGET:
      HintNode = M->getOrInsertNamedMetadata("visc_hint_spir");
      break;
    case visc::CUDNN_TARGET:
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cudnn");
      break;
    case visc::PROMISE_TARGET:
      HintNode = M->getOrInsertNamedMetadata("visc_hint_promise");
      break;
    case visc::CPU_OR_GPU_TARGET:
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu_gpu");
      break;
    case visc::CPU_OR_SPIR_TARGET:
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu_spir");
      break;
    case visc::CPU_TARGET:
      HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu");
      break;
    default:
      llvm_unreachable("Unsupported Target Hint!");
      break;
  }

  // Gather metadata nodes, and keep those not associated with this function
  MDNode* N = MDNode::get(M->getContext(),
                          ArrayRef<Metadata*>(ValueAsMetadata::get(F)));
  std::vector<MDNode*> MDNodes;

  for (unsigned i = 0; i < HintNode->getNumOperands(); i++) {
    MDNode* MDN = HintNode->getOperand(i);
    if (MDN == N) {
      continue;
    }
    MDNodes.push_back(MDN);
  }

  HintNode->dropAllReferences();

  for (unsigned i = 0; i < MDNodes.size(); i++) {
    HintNode->addOperand(MDNodes[i]);
  }

}

visc::Target getPreferredTarget(Function* F) {
  DEBUG(errs() << "Finding preferred target for " << F->getName() << "\n");
  Module* M = F->getParent();
  NamedMDNode* HintNode = M->getOrInsertNamedMetadata("visc_hint_gpu");
  for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
    MDNode* N = HintNode->getOperand(i);
    Value* FHint = dyn_cast<ValueAsMetadata>(N->getOperand(0).get())->getValue();
    if(F == FHint)
      return visc::GPU_TARGET;
  }

  HintNode = M->getOrInsertNamedMetadata("visc_hint_spir");
  for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
    MDNode* N = HintNode->getOperand(i);
    Value* FHint = dyn_cast<ValueAsMetadata>(N->getOperand(0).get())->getValue();
    if(F == FHint)
      return visc::SPIR_TARGET;
  }

  HintNode = M->getOrInsertNamedMetadata("visc_hint_cudnn");
  for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
    MDNode* N = HintNode->getOperand(i);
    Value* FHint = dyn_cast<ValueAsMetadata>(N->getOperand(0).get())->getValue();
    if(F == FHint)
      return visc::CUDNN_TARGET;
  }

  HintNode = M->getOrInsertNamedMetadata("visc_hint_promise");
  for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
    MDNode* N = HintNode->getOperand(i);
    Value* FHint = dyn_cast<ValueAsMetadata>(N->getOperand(0).get())->getValue();
    if(F == FHint)
      return visc::PROMISE_TARGET;
  }

  HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu_gpu");
  for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
    MDNode* N = HintNode->getOperand(i);
    Value* FHint = dyn_cast<ValueAsMetadata>(N->getOperand(0).get())->getValue();
    if(F == FHint)
      return visc::CPU_OR_GPU_TARGET;
  }

  HintNode = M->getOrInsertNamedMetadata("visc_hint_cpu_spir");
  for(unsigned i = 0; i < HintNode->getNumOperands(); i++) {
    MDNode* N = HintNode->getOperand(i);
    Value* FHint = dyn_cast<ValueAsMetadata>(N->getOperand(0).get())->getValue();
    if(F == FHint)
      return visc::CPU_OR_SPIR_TARGET;
  }
  return visc::CPU_TARGET;
}


} // End of namespace

#endif //VISC_UTILS_HEADER
